/*	Serpent Strong Encryption Plugin for Back Orifice 2000
	Copyright (C) 1999, Daniel Roethlisberger

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	The author of this program may be contacted at admin@roe.ch.
*/

#include <windows.h>
#include <encryption.h>
#include <config.h>
#include <time.h>
#include <sys/timeb.h>
#include "bo_serpent.h"
#include "hash.h"
#include "serpent.h"
#include "myserpent.h"


ENCRYPTION_ENGINE g_SerpentEngine = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};

// called when engine is inserted into handler
int __cdecl Serpent_Insert(void)
{
	if(GetCfgStr(g_szSerpentEncryptOptions, SERPENT_CONFIG_KEY_STRING) == NULL)
		return -1;
	return 0;
}

// called when engine is removed from handler
int __cdecl Serpent_Remove(void)
{
	return 0;
}

// returns the ID string of the engine
char * __cdecl Serpent_Query(void)
{
	return SERPENT_QUERY_STRING;
}

// called before any key or encryption takes place to initialise
void * __cdecl Serpent_Startup(void)
{
	// read the configuration
	char *svKey = GetCfgStr(g_szSerpentEncryptOptions, SERPENT_CONFIG_KEY_STRING);
	if(svKey == NULL)
		return NULL;
	if(svKey[0] == '\0')
		return NULL;

	// allocate a context structure
	SERPENT_CTX *data;
	data = (SERPENT_CTX*)malloc(sizeof(SERPENT_CTX));
	if(data == NULL)
		return NULL;

	// set the configuration
	Serpent_SetKey((void*)data, svKey);

	return data;
}

// called after encryption has been finished to clean up
int __cdecl Serpent_Shutdown(void *pInternal)
{
	SERPENT_CTX *data = (SERPENT_CTX*)pInternal;
	free(data);
	return 0;
}

// called to set key string
int __cdecl Serpent_SetKey(void *pInternal, char *svKey)
{
	SERPENT_CTX *data = (SERPENT_CTX*)pInternal;

	// hash the key string to a 256 bit user key
	HASH_CTX* ctx = (HASH_CTX*)malloc(sizeof(HASH_CTX));
	HashInit(ctx);
	HashUpdate(ctx, (BYTE*)svKey, strlen(svKey));
	HashFinal((BYTE*)data->SerpentUserKey, ctx);
	free(ctx);

	strcpy(data->svSerpentKey, svKey);

	// let the Serpent engine initialise the key context
	SerpentKeyInit(data->SerpentUserKey, data->SerpentKey);

	return 0;
}

// returns key string
char * __cdecl Serpent_GetKey(void *pInternal)
{
	SERPENT_CTX *data = (SERPENT_CTX*)pInternal;
	return data->svSerpentKey;
}

// returns encrypted buffer that must be freed
BYTE * __cdecl Serpent_Encrypt(void *pInternal, BYTE *pBuffer, int nBufLen, int *pnOutBufLen)
{
	SERPENT_CTX *data = (SERPENT_CTX*)pInternal;

	BYTE *buf;
	int blocks, baselen, leftover;

	// init c standard randomiser with millisecs since last startup
	// not a truly random number sequence, but good enough for this purpuse
	srand(GetTickCount());

	// check on length (must be a multiple of 128 bits)
	leftover = nBufLen & 15;
	baselen = nBufLen + 16 - leftover;

	// add padding
	buf = (BYTE*)malloc(baselen + 16);

	for(int g = 0; g < 15 - leftover; g++)
		buf[16+nBufLen+g] = rand() % 256;

	buf[baselen+15] |= 0x0F;
	buf[baselen+15] &= (unsigned char)(15-leftover); // 16-leftover-1 'cos we only got 4 bit

	// generate IV
	for(int h = 0; h < 16; h++)
		data->SerpentInitVect[h] = rand() % 256;
	
	memcpy(buf, data->SerpentInitVect, 16);
	memcpy(buf+16, pBuffer, nBufLen);

	// number of blocks to encode
	blocks = baselen >> 4;

	// CBC Mode: cypherblock = enc(plainblock XOR lastcypherblock)
	for(int i = 0; i < blocks; i++)
	{
		for(int j = 0; j < SERPENT_BLOCK_SIZE; j++)
		{
			buf[16+i*SERPENT_BLOCK_SIZE+j] ^= data->SerpentInitVect[j];
		}
		SerpentEncrypt((u4byte*)&buf[16+i*SERPENT_BLOCK_SIZE], data->SerpentKey);
		memcpy(data->SerpentInitVect, &buf[16+i*SERPENT_BLOCK_SIZE], SERPENT_BLOCK_SIZE);
	}
	*pnOutBufLen = baselen + 16;
	
	return buf;
}

// returns decrypted buffer that must be freed
BYTE * __cdecl Serpent_Decrypt(void *pInternal, BYTE *pBuffer, int nBufLen, int *pnOutBufLen)
{
	SERPENT_CTX *data = (SERPENT_CTX*)pInternal;

	BYTE *buf;
	int blocks = 0;
	BYTE NewInitVect[SERPENT_BLOCK_SIZE] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

	// check on length (must be a multiple of 128 bits)
	if(nBufLen & 15)
		return NULL;
	nBufLen -= 16;

	memcpy(data->SerpentInitVect, pBuffer, SERPENT_BLOCK_SIZE);

	buf = (BYTE*)malloc(nBufLen);
	
	memcpy(buf, pBuffer+16, nBufLen);

	// number of blocks to encode
	blocks = nBufLen >> 4;

	// CBC Mode: plainblock = dec(cypherblock) XOR lastcypherblock
	for(int i = 0; i < blocks; i++)
	{
		memcpy(NewInitVect, &buf[i*SERPENT_BLOCK_SIZE], SERPENT_BLOCK_SIZE);
		SerpentDecrypt((u4byte*)&buf[i*SERPENT_BLOCK_SIZE], data->SerpentKey);
		for(int j = 0; j < SERPENT_BLOCK_SIZE; j++)
		{
			buf[i*SERPENT_BLOCK_SIZE+j] ^= data->SerpentInitVect[j];
		}
		memcpy(data->SerpentInitVect, NewInitVect, SERPENT_BLOCK_SIZE);
	}
	
	// remove padding
	nBufLen -= (buf[nBufLen-1] & 0x0F) + 1; // + 1 'cos we only got 4 bit

	*pnOutBufLen = nBufLen;

	nBufLen += 16;

	return buf;
}

// generate new keys, we just get them from the config string
int __cdecl Serpent_CreateNewKeys(void *pInternal)
{
	char *svKey = GetCfgStr(g_szSerpentEncryptOptions, SERPENT_CONFIG_KEY_STRING);
	if(svKey == NULL)
		return -1;
	if(svKey[0] == '\0')
		return -1;

	SERPENT_CTX *data = (SERPENT_CTX*)pInternal;
	free(data);
	data = (SERPENT_CTX*)malloc(sizeof(SERPENT_CTX));
	if(data == NULL)
		return -1;
	
	Serpent_SetKey((void*)data, svKey);

	return 0;
}

// free buffers returned by Serpent_Encrypt or Serpent_Decrypt
void __cdecl Serpent_Free(void *pInternal, BYTE *pBuffer)
{
	free(pBuffer);
}

// returns a struct containing the function addresses
ENCRYPTION_ENGINE *GetSerpentEncryptionEngine(void)
{
	g_SerpentEngine.pInsert = Serpent_Insert;
	g_SerpentEngine.pRemove = Serpent_Remove;
	g_SerpentEngine.pQuery = Serpent_Query;
	g_SerpentEngine.pStartup = Serpent_Startup;
	g_SerpentEngine.pShutdown = Serpent_Shutdown;
	g_SerpentEngine.pSetEncryptKey = Serpent_SetKey;
	g_SerpentEngine.pSetDecryptKey = Serpent_SetKey;
	g_SerpentEngine.pGetEncryptKey = Serpent_GetKey;
	g_SerpentEngine.pGetDecryptKey = Serpent_GetKey;
	g_SerpentEngine.pEncrypt = Serpent_Encrypt;
	g_SerpentEngine.pDecrypt = Serpent_Decrypt;
	g_SerpentEngine.pCreateNewKeys = Serpent_CreateNewKeys;
	g_SerpentEngine.pFree = Serpent_Free;
	
	return &g_SerpentEngine;
}


